/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */


#ifndef CIPHER_H
#define CIPHER_H 1

# ifdef LIBFLUX_BUILD
#  include "types.h"
#  include "mpi.h"
# else
#  include <flux/types.h>
#  include <flux/mpi.h>
# endif


typedef enum
{
  CIPHER_NONE = 0, CIPHER_ANY, CIPHER_TWOFISH, CIPHER_ELGAMAL
} CIPHER_TYPE;


/* The cipher context structure. */

typedef struct _cipher
{
  u16 sym_type, asym_type;  /* Symmetric/asymmetric schemes used */

  void *sym_key_encrypt;    /* Symmetric key used to encrypt this session */
  void *sym_key_decrypt;    /* Symmetric key used to decrypt this session */

  void *asym_key_encrypt;   /* Asymmetric key used to encrypt (her public) */
  void *asym_key_decrypt;   /* Asymmetric key used to decrypt (my secret) */

  void *asym_key_sign;      /* Asymmetric key used to sign (my secret) */
  void *asym_key_verify;    /* Asymmetric key used to verify (her public) */
}
CIPHER;


/* --- Generic cipher functions --- */

CIPHER *cipher_new(CIPHER_TYPE sym, CIPHER_TYPE asym);
CIPHER *cipher_dup(CIPHER *old);
void cipher_del(CIPHER *cip);
void cipher_gen(CIPHER *cip);

int cipher_import_asym(CIPHER *cip, byte *in);
void cipher_export_asym(CIPHER *cip, byte **out, unsigned int *len);

int cipher_import_sym(CIPHER *cip, byte *in);
void cipher_export_sym(CIPHER *cip, byte **out, unsigned int *len);

void cipher_permute_in(CIPHER *cip);
void cipher_permute_out(CIPHER *cip);

int cipher_has_keys(CIPHER *cip);

#define cipher_set_sym_key_encrypt(cip, type, key) (cip)->sym_key_encrypt = (key)
#define cipher_set_sym_key_decrypt(cip, type, key) (cip)->sym_key_decrypt = (key)

#define cipher_set_asym_key_encrypt(cip, type, key) (cip)->asym_key_encrypt = (key)
#define cipher_set_asym_key_decrypt(cip, type, key) (cip)->asym_key_decrypt = (key)

#define cipher_set_asym_key_sign(cip, type, key) (cip)->asym_key_sign = (key)
#define cipher_set_asym_key_verify(cip, type, key) (cip)->asym_key_verify = (key)


/* Structure for an expanded Twofish key.  s contains the key-dependent
 * S-boxes composed with the MDS matrix; w contains the eight "whitening"
 * subkeys, K[0] through K[7]. k holds the remaining, "round" subkeys.  Note
 * that k[i] corresponds to what the Twofish paper calls K[i+8]. */

typedef struct
{
  u16 key_len;    /* Key length in bytes */
  u16 v_pos;

  union
  {
    byte key[32];   /* Unexpanded key */
    unsigned long __align;
  }
  key;

  union
  {
    byte v[16];     /* Vectors (not secret) */
    unsigned long __align;
  }
  v;

  u32 s[4][256],
      w[8],
      k[32];
}
CIPHER_TWOFISH_KEY;


/* Structures for ElGamal secret and public keys. Note that the public key
 * is just a subset of the secret one. */

typedef struct
{
  MPI p;  /* Prime (may be shared between users) */
  MPI g;  /* Group generator (may be shared between users) */
  MPI y;  /* g^x mod p */
  MPI x;  /* Exponent (the secret part) */
}
CIPHER_ELGAMAL_SKEY;


typedef struct
{
  MPI p;  /* Prime (may be shared between users) */
  MPI g;  /* Group generator (may be shared between users) */
  MPI y;  /* g^x mod p */
}
CIPHER_ELGAMAL_PKEY;


/* --- Prime stuff --- */

extern int _prime_small_num;
extern unsigned short _prime_small[];


/* --- Prototypes --- */

/* RMD160 hashing */

void rmd160_hash(unsigned char *out, const unsigned char *in, size_t len);

/* Twofish symmetric cipher */

CIPHER_TWOFISH_KEY *twofish_gen(unsigned nbits);
CIPHER_TWOFISH_KEY *twofish_dup(CIPHER_TWOFISH_KEY *key);
void twofish_del(CIPHER_TWOFISH_KEY *key);

CIPHER_TWOFISH_KEY *twofish_import(const byte *key, const unsigned keylen);
void twofish_export(CIPHER_TWOFISH_KEY *key, byte **out, unsigned int *len);

void twofish_import_iv(CIPHER_TWOFISH_KEY *key, const byte *iv);
void twofish_export_iv(CIPHER_TWOFISH_KEY *key, byte **out, unsigned int *len);
void twofish_permute_iv(CIPHER_TWOFISH_KEY *key);

#define twofish_bits(key) ((key)->key_len << 3)  /* Returns bits */
#define twofish_len(key) ((key)->key_len)  /* Returns bytes */
void twofish_set(CIPHER_TWOFISH_KEY *ctx, const byte *key, const unsigned len);

void twofish_encrypt(const CIPHER_TWOFISH_KEY *ctx, byte *out, const byte *in);
void twofish_decrypt(const CIPHER_TWOFISH_KEY *ctx, byte *out, const byte *in);

void twofish_encrypt_cbc(CIPHER_TWOFISH_KEY *key, byte *out, const byte *in);
void twofish_decrypt_cbc(CIPHER_TWOFISH_KEY *key, byte *out, const byte *in);

void twofish_encrypt_cfb(CIPHER_TWOFISH_KEY *key, byte *out, const byte *in, size_t len);
void twofish_decrypt_cfb(CIPHER_TWOFISH_KEY *key, byte *out, const byte *in, size_t len);

void twofish_encrypt_ofb(CIPHER_TWOFISH_KEY *key, byte *out, const byte *in, size_t len);
void twofish_decrypt_ofb(CIPHER_TWOFISH_KEY *key, byte *out, const byte *in, size_t len);

/* ElGamal asymmetric cipher */

/* Secret key manipulation */

CIPHER_ELGAMAL_SKEY *elgamal_gen_secret(unsigned nbits);
CIPHER_ELGAMAL_SKEY *elgamal_dup_secret(CIPHER_ELGAMAL_SKEY *skey);
CIPHER_ELGAMAL_SKEY *elgamal_import_secret(const byte *in);
void elgamal_export_secret(CIPHER_ELGAMAL_SKEY *skey, byte **out, unsigned int *len);
int elgamal_bits_secret(CIPHER_ELGAMAL_SKEY *skey);
int elgamal_test_secret(CIPHER_ELGAMAL_SKEY *skey);
void elgamal_del_secret(CIPHER_ELGAMAL_SKEY *skey);

/* Public key manipulation */

CIPHER_ELGAMAL_PKEY *elgamal_gen_public(CIPHER_ELGAMAL_SKEY *skey);
CIPHER_ELGAMAL_PKEY *elgamal_dup_public(CIPHER_ELGAMAL_PKEY *pkey);
CIPHER_ELGAMAL_PKEY *elgamal_import_public(const byte *in);
void elgamal_export_public(CIPHER_ELGAMAL_PKEY *pkey, byte **out, unsigned int *len);
int elgamal_bits_public(CIPHER_ELGAMAL_PKEY *pkey);
void elgamal_del_public(CIPHER_ELGAMAL_PKEY *pkey);

/* Mixed-key manipulation */

int elgamal_is_pair(CIPHER_ELGAMAL_PKEY *pkey, CIPHER_ELGAMAL_SKEY *skey);
#define elgamal_bits(key) mpi_get_nbits(((CIPHER_ELGAMAL_SKEY *) key)->p)
#define elgamal_block_size(key) ((elgamal_bits(key) + 3) >> 2)  /* (x / 8) * 2 */

/* Encryption, decryption, signing and verification */

void elgamal_encrypt(CIPHER_ELGAMAL_PKEY *pkey,
                     void **cipher, size_t *cipher_len,
                     void *plain, size_t plain_len);
void elgamal_decrypt(CIPHER_ELGAMAL_SKEY *skey,
                     void **plain, size_t *plain_len,
                     void *cipher, size_t cipher_len);

void elgamal_sign(CIPHER_ELGAMAL_SKEY *skey,
                  void **sign, size_t *sign_len,
                  void *hash, size_t hash_len);
int elgamal_verify(CIPHER_ELGAMAL_PKEY *pkey,
                   void *sign, size_t sign_len,
                   void *hash, size_t hash_len);

#endif  /* CIPHER_H */
